import copy
from dataclasses import dataclass
import os
import queue
import re
import threading
import time
import urllib.parse as urlparse
from typing import Optional
import concurrent.futures

from bs4 import BeautifulSoup
import requests

from dokuWikiDumper.utils.util import smkdirs, uopen
from dokuWikiDumper.utils.util import print_with_lock as print
from dokuWikiDumper.utils.config import runtime_config

exit_event = threading.Event()

@dataclass
class DumpMediaParams:
    dump_dir: str
    title: str
    title_index: int
    base_url: str
    session: requests.Session
    fetch_url: str


def getFiles(url, ns: str = '',  dumpDir: str = '', session: requests.Session=None):
    """ Return a list of media filenames of a wiki """

    if dumpDir and os.path.exists(dumpDir + '/dumpMeta/files.txt'):
        with uopen(dumpDir + '/dumpMeta/files.txt', 'r') as f:
            files = f.read().splitlines()
            if files[-1] == '--END--':
                print('Loaded %d files from %s' %
                      (len(files) - 1, dumpDir + '/dumpMeta/files.txt'))
                return files[:-1]  # remove '--END--'

    files = set()
    ajax = urlparse.urljoin(url, 'lib/exe/ajax.php')
    medialist = BeautifulSoup(
        session.post(ajax, {
            'call': 'medialist',
            'ns': ns,
            'do': 'media'
        }).text, runtime_config.html_parser)
    medians = BeautifulSoup(
        session.post(ajax, {
            'call': 'medians',
            'ns': ns,
            'do': 'media'
        }).text, runtime_config.html_parser)
    imagelinks = medialist.find_all(
        'a',
        href=lambda x: x and re.findall(
            '[?&](media|image)=',
            x))
    for a in imagelinks:
        query = urlparse.parse_qs(urlparse.urlparse(a['href']).query)
        key = 'media' if 'media' in query else 'image'
        files.add(query[key][0])
    files = list(files)
    namespacelinks = medians.find_all('a', {'class': 'idx_dir', 'href': True})
    for a in namespacelinks:
        query = urlparse.parse_qs(urlparse.urlparse(a['href']).query)
        files += getFiles(url, query['ns'][0], session=session)
    print('Found %d files in namespace %s' % (len(files), ns or '(all)'))

    if dumpDir:
        smkdirs(dumpDir + '/dumpMeta')
        with uopen(dumpDir + '/dumpMeta/files.txt', 'w') as f:
            f.write('\n'.join(files))
            f.write('\n--END--\n')

    return files


def dump_media(*, base_url: str, dumpDir: str, session: requests.Session, threads: int = 1, ignore_errors: bool = False):

    smkdirs(dumpDir + '/media')

    fetch = urlparse.urljoin(base_url, 'lib/exe/fetch.php')
    # media_repo = urlparse.urljoin(base_url, '_media')

    files = getFiles(base_url, dumpDir=dumpDir, session=session)
    tasks_queue: queue.Queue[DumpMediaParams] = queue.Queue(maxsize=threads)

    task_templ = DumpMediaParams(dump_dir=dumpDir, base_url=base_url, session=session, fetch_url=fetch,
                                title_index=-999, title="dokuwikidumper_placehold")

    def task_generator():
        for index, title in enumerate(files):
            task = copy.copy(task_templ)
            task.title_index = index
            task.title = title
            tasks_queue.put(task)
            print('Media: (%d/%d): [[%s]] ...' % (index+1, len(files), title))

            if exit_event.is_set():
                print('task generator exit (exit event set)')
                return

        tasks_queue.join()
        print('All tasks done, terminating workers...')
        exit_event.set()

    tg_thread = threading.Thread(target=task_generator, name='task-generator')
    tg_thread.daemon = True
    tg_thread.start()

    def _dump_media_action(task: DumpMediaParams, ignore_errors: bool):
        try:
            download_media_file(task)
        except Exception as e:
            if not ignore_errors:
                raise e
            else:
                print('[',task.title_index+1,'] Error in sub thread: (', e, ') ignored')

    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        futures = set()
        while not exit_event.is_set():
            try:
                task = tasks_queue.get(timeout=1)
            except queue.Empty:
                continue
            f = executor.submit(_dump_media_action, task, ignore_errors)
            f.add_done_callback(lambda f: tasks_queue.task_done())
            futures.add(f)

            if len(futures) >= threads:
                _done, futures = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
                for f in _done:
                    f.result()

        while futures:
            _done, futures = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
            for f in _done:
                f.result()

    tg_thread.join()


def download_media_file(task: DumpMediaParams):
    child_path = task.title.replace(':', '/')
    child_path = child_path.lstrip('/')
    child_path = '/'.join(child_path.split('/')[:-1])
    smkdirs(task.dump_dir + '/media/' + child_path)
    file = task.dump_dir + '/media/' + task.title.replace(':', '/')
    local_size = -1
    if os.path.exists(file):
        local_size = os.path.getsize(file)
    with task.session.get(task.fetch_url, params={'media': task.title},
                    stream=True, headers={'Referer': task.base_url}
                    ) as r:
        r.raise_for_status()

        if local_size == -1:  # file does not exist
            to_download = True
        else:
            remote_size = int(r.headers.get('Content-Length', -2))
            if local_size == remote_size:  # file exists and is complete
                print('[%d] File [[%s]] exists (%d bytes)' % (task.title_index+1, task.title, local_size))
                to_download = False
            elif remote_size == -2:
                print('[%d] File [[%s]] cannot get remote size ("Content-Length" missing), ' % (task.title_index+1, task.title) +
                      'will re-download anyway')
                to_download = True
            else:
                to_download = True  # file exists but is incomplete

        if to_download:
            with open(file, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
                print('[%d] File [[%s]] Done' % (task.title_index+1, task.title))
        else:
            r.close()

        # modify mtime based on Last-Modified header
        last_modified = r.headers.get('Last-Modified', None)
        if last_modified:
            mtime = time.mktime(time.strptime(
                last_modified, '%a, %d %b %Y %H:%M:%S %Z'))
            atime = os.stat(file).st_atime
            # atime is not modified
            os.utime(file, times=(atime, mtime))



import copy
from dataclasses import dataclass
import queue
import time
import threading
from typing import Callable
import concurrent.futures

from requests import Session

from dokuWikiDumper.exceptions import ActionEditDisabled, ActionEditTextareaNotFound, DispositionHeaderMissingError

from .revisions import get_revisions, get_source_edit, get_source_export, save_page_changes
from .titles import load_get_save_titles
from dokuWikiDumper.utils.util import smkdirs, uopen
from dokuWikiDumper.utils.util import print_with_lock as print


@dataclass
class DumpPageParams:
    dump_dir: str
    get_source: Callable
    title_index: int
    title: str
    doku_url: str
    session: Session
    current_only: bool

exit_event = threading.Event()

def dump_content(*, doku_url: str, dump_dir: str, session: Session, threads: int = 1,
                 ignore_errors: bool = False, ignore_action_disabled_edit: bool = False, current_only: bool = False):
    titles = load_get_save_titles(dump_dir=dump_dir, url=doku_url, session=session)

    if not len(titles):
        print('Empty wiki')
        return False

    r1 = session.get(doku_url, params={'id': titles[0], 'do': 'export_raw'})

    get_source = get_source_export
    if 'html' in r1.headers['content-type']:
        print('\nWarning: export_raw action not available, using edit action\n')
        time.sleep(3)
        get_source = get_source_edit

    task_templ = DumpPageParams(dump_dir=dump_dir, doku_url=doku_url, session=session, get_source=get_source, current_only=current_only,
                          title_index=-999, title="dokuwikidumper_placehold")

    tasks_queue: queue.Queue[DumpPageParams] = queue.Queue(maxsize=threads)

    def task_generator():
        for index, title in enumerate(titles):
            task = copy.copy(task_templ)
            task.title_index = index
            task.title = title
            tasks_queue.put(task)
            print('Content: (%d/%d): [[%s]] ...' % (index+1, len(titles), title))

            if exit_event.is_set():
                print('task generator exit (exit event set)')
                return

        tasks_queue.join()
        print('All tasks done, terminating workers...')
        exit_event.set()

    tg_thread = threading.Thread(target=task_generator, name='task-generator')
    tg_thread.daemon = True
    tg_thread.start()

    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        futures = set()
        while not exit_event.is_set():
            try:
                task = tasks_queue.get(timeout=1)
            except queue.Empty:
                continue
            f = executor.submit(_dump_action, task, ignore_errors, ignore_action_disabled_edit)
            f.add_done_callback(lambda f: tasks_queue.task_done())
            futures.add(f)

            if len(futures) >= threads:
                _done, futures = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
                for f in _done:
                    f.result()

        while futures:
            _done, futures = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
            for f in _done:
                f.result()


    tg_thread.join()


def _dump_action(task: DumpPageParams, ignore_errors: bool, ignore_action_disabled_edit: bool):
    try:
        dump_page(task)
    except Exception as e:
        if isinstance(e, ActionEditDisabled) and ignore_action_disabled_edit:
            print('[',task.title_index,'] action disabled: edit. ignored')
        elif isinstance(e, ActionEditTextareaNotFound) and ignore_action_disabled_edit:
            print('[',task.title_index,'] action edit: textarea not found. ignored')
        elif not ignore_errors:
            raise e
        else:
            print('[',task.title_index,'] Error in sub thread: (', e, ') ignored')


def dump_page(task: DumpPageParams):
    source = task.get_source(task.doku_url, task.title, session=task.session)
    msg_header = '['+str(task.title_index + 1)+']: '
    child_path = task.title.replace(':', '/')
    child_path = child_path.lstrip('/')
    child_path = '/'.join(child_path.split('/')[:-1])

    smkdirs(task.dump_dir, '/pages/' + child_path)
    with uopen(task.dump_dir + '/pages/' + task.title.replace(':', '/') + '.txt', 'w') as f:
        f.write(source)

    if task.current_only:
        print(msg_header, '    [[%s]] saved.' % (task.title))
        return

    revs = get_revisions(task.doku_url, task.title, session=task.session, msg_header=msg_header)


    save_page_changes(dumpDir=task.dump_dir, child_path=child_path, title=task.title, 
                       revs=revs, msg_header=msg_header)

    for rev in revs[1:]:
        if 'id' in rev and rev['id']:
            try:
                txt = task.get_source(task.doku_url, task.title, rev['id'], session=task.session)
                smkdirs(task.dump_dir, '/attic/' + child_path)
                with uopen(task.dump_dir + '/attic/' + task.title.replace(':', '/') + '.' + rev['id'] + '.txt', 'w') as f:
                    f.write(txt)
                print(msg_header, '    Revision %s of [[%s]] saved.' % (
                    rev['id'], task.title))
            except DispositionHeaderMissingError:
                print(msg_header, '    Revision %s of [[%s]] is empty. (probably deleted)' % (
                    rev['id'], task.title))
        else:
            print(msg_header, '    Revision %s of [[%s]] failed: %s' % (rev['id'], task.title, 'Rev id not found (please check ?do=revisions of this page)'))



import os
import traceback

from dokuWikiDumper.version import get_version


class VersionOutdatedError(Exception):
    def __init__(self, version):
        self.version = version

    def __str__(self):
        return "Version outdated: %s" % self.version

class HTTPStatusError(Exception):
    def __init__(self, r):
        self.status_code = r.status_code
        self.url = r.url

    def __str__(self):
        return "HTTP Status Code: %s, URL: %s" % (self.status_code, self.url)


class DispositionHeaderMissingError(Exception):
    def __init__(self, r):
        self.url = r.url

    def __str__(self):
        return "Disposition header missing, URL: %s" % self.url

class ActionExportRawDisabled(Exception):
    def __init__(self, title):
        self.title = title

    def __str__(self):
        return "Action: export_raw is disabled for [[%s]]" % self.title

class ActionEditDisabled(Exception):
    def __init__(self, title):
        self.title = title

    def __str__(self):
        return "Action: edit is disabled for [[%s]]" % self.title


class ActionIndexDisabled(Exception):
    def __init__(self):
        pass

    def __str__(self):
        return "Action: index is disabled"
    

class ActionEditTextareaNotFound(Exception):
    def __init__(self, title):
        self.title = title

    def __str__(self):
        return "Action: edit: textarea not found for [[%s]]" % self.title


class ActionExportHtmlDisabled(Exception):
    def __init__(self, title):
        self.title = title

    def __str__(self):
        return "Action: export_xhtml is disabled for [[%s]]" % self.title


class ActionRevisionsDisabled(Exception):
    def __init__(self, title):
        self.title = title

    def __str__(self):
        return "Action: revisions is disabled for [[%s]]" % self.title


class RevisionListNotFound(Exception):
    def __init__(self, title):
        self.title = title

    def __str__(self):
        return "Revision list not found for [[%s]]" % self.title


def show_edge_case_warning(**context):
    if os.environ.get('EDGECASE_OK'):
        return

    print(
    "[WARNING]\n"
    "--------------------------------------------\n"
    "The program is about to enter an edge case code, "
    "which lacks real world testing, I'm not sure if the code that runs next can handle it properly. "
    "So I hope you could paste the following details to <https://github.com/saveweb/dokuwiki-dumper/discussions> "
    "to help me improve the code, Thanks!")
    print("------------------------------------------")
    calledfrom = traceback.extract_stack(limit=2)[0]
    print("VERSION:", get_version())
    print("FUNC:", f'{calledfrom.filename}:{calledfrom.lineno} ', "FUNC:", calledfrom.name)
    print("CONTEXT:", context)
    print("------------------------------------------")
    print("To continue executing the edge case code, re-run with environment variable EDGECASE_OK=1 set.")
    os._exit(13)



__VERSION__ = 'unknown'

try:
    from importlib.metadata import version
    __VERSION__ = version('dokuwikidumper')
except Exception:
    pass

def get_version():
    return __VERSION__


def get_latest_version():
    '''Returns the latest version of dokuwikidumper.'''
    project_url_pypi = 'https://pypi.org/pypi/dokuwikidumper/json'
    project_url_github = 'https://api.github.com/repos/saveweb/dokuwiki-dumper/releases/latest'
    
    import requests
    try:
        response = requests.get(project_url_pypi, timeout=5, headers={'Accept': 'application/json', 'Accept-Encoding': 'gzip'})
    except requests.exceptions.Timeout or requests.exceptions.ConnectionError:
        print('Warning: Could not get latest version of dokuwikidumper from pypi.org. (Timeout)')
        return None
    if response.status_code == 200:
        data = response.json()
        latest_version = data['info']['version']
        return latest_version
    else:
        print('Warning: Could not get latest version of dokuwikidumper.')
        return None



from datetime import datetime
import html
import os
import re
from ipaddress import ip_address, IPv4Address, IPv6Address
import time
import urllib.parse as urlparse
import logging

from typing import List, Optional, TypedDict

import requests
from bs4 import BeautifulSoup, Tag

from dokuWikiDumper.exceptions import ActionEditDisabled, ActionEditTextareaNotFound, ActionRevisionsDisabled, DispositionHeaderMissingError, HTTPStatusError, RevisionListNotFound, show_edge_case_warning
from dokuWikiDumper.utils.util import check_int, print_with_lock as print, smkdirs, uopen
from dokuWikiDumper.utils.config import runtime_config


logger = logging.getLogger(__name__)

# args must be same as get_source_edit(), even if not used
def get_source_export(url: str, title: str, rev: str = '', *, session: requests.Session):
    """Export the raw source of a page (at a given revision)"""

    r = session.get(url, params={'id': title, 'rev': rev, 'do': 'export_raw'})
    if r.status_code != 200:
        raise HTTPStatusError(r)

    disposition_ok = 'Content-Disposition' in r.headers
    content_type_ok = 'text/plain' in r.headers.get('content-type', '')
    if disposition_ok or content_type_ok:
        return r.text

    raise DispositionHeaderMissingError(r)


# args must be same as get_source_export(), even if not used
def get_source_edit(url: str, title: str, rev: str = '', *, session: requests.Session):
    """Export the raw source of a page by scraping the edit box content. Yuck."""

    r = session.get(url, params={'id': title, 'rev': rev, 'do': 'edit'})
    soup = BeautifulSoup(r.text, runtime_config.html_parser)
    source = None

    textarea = soup.find('textarea', {'name': 'wikitext'})
    if textarea:
        source = ''.join(textarea.text).strip()
    elif 'Action disabled: source' in r.text:
        raise ActionEditDisabled(title)
    else:
        raise ActionEditTextareaNotFound(title)

    return source

class Revision(TypedDict):
    """ (None if not found or failed) """
    id: Optional[str]
    user: Optional[str]
    sum: Optional[str]
    date: Optional[str]
    minor: bool
    """ default: False """
    sizechange: int
    """ default: 0 """

def get_revisions(doku_url, title: str, session: requests.Session, msg_header: str = '')->List[Revision]:
    """ Get the revisions of a page. This is nontrivial because different versions of DokuWiki return completely different revision HTML.

    Returns a dict with the following keys:
    """

    revs: List[Revision] = []
    rev_tmplate: Revision = {
        'id': None, # str(int)
        'user': None, # str
        'sum': None, # str
        'date': None, # str
        'minor': False, # bool
        'sizechange': 0,
    }

    i = 0
    continue_index = -1
    cont = True

    while cont:
        r = session.get(
            doku_url,
            params={
                'id': title,
                'do': 'revisions',
                'first': str(continue_index)})

        soup = BeautifulSoup(r.text, runtime_config.html_parser)

        lis = None

        # check if form#page__revisions exists
        if page__revisions := soup.find('form', {'id': 'page__revisions'}):
            logger.debug('page__revisions: %s', page__revisions)
            if ul := page__revisions.find('ul'):
                assert isinstance(ul, Tag)
                lis = ul.find_all('li')

        # outdate dokuwiki version? try another way.
        if div_page := soup.find('div', {'class': 'page'}):
            logger.debug('div.page: %s', div_page)
            if ul := div_page.find('ul'):
                assert isinstance(ul, Tag)
                lis = ul.find_all('li')

        if lis is None:
            if err_msg := soup.find('div', {'class': 'error'}):
                if 'Action disabled: revisions' in err_msg.text:
                    raise ActionRevisionsDisabled(title)

            raise RevisionListNotFound(title)

        assert lis is not None

        for li in lis:
            li: Tag
            rev = {}

            checkbox = li.find('input', {'type': 'checkbox'})
            rev_hrefs = li.find_all(
                'a', href=lambda href: isinstance(href, str) and (
                    '&rev=' in href or '?rev=' in href))

            # id: optional(str(id)): rev_id, not title name.
            if checkbox:
                rev['id'] = check_int(checkbox.get('value', None))

            if rev_hrefs and rev.get('id', None) is None:
                obj1 = rev_hrefs[0]['href']
                obj2 = urlparse.urlparse(obj1).query
                obj3 = urlparse.parse_qs(obj2)
                if 'rev' in obj3:
                    rev['id'] = check_int(obj3['rev'][0])
                else:
                    rev['id'] = None
                del (obj1, obj2, obj3)

            # use_hidden_rev
            if rev.get('id', None) is None:
                obj1 = li.find('input', {'type': 'hidden'})
                if obj1 is not None and 'value' in obj1:
                    rev['id'] = check_int(obj1['value'])
                del (obj1)

            # minor: bool
            rev['minor'] = li.has_attr('class') and 'minor' in li['class']

            # summary: optional(str)
            sum_span = li.find_all('span', {'class': 'sum'})
            if sum_span:
                sum_span = sum_span[0]
                sum_text = sum_span.text.split(' ')[1:]
                if sum_span.find_all('bdi'):
                    rev['sum'] = html.unescape(
                        sum_span.find('bdi').text).strip()
                else:
                    rev['sum'] = html.unescape(' '.join(sum_text)).strip()
            else:
                print(msg_header, '    ', repr(
                    li.text).replace('\\n', ' ').strip())
                wikilink1 = li.find('a', {'class': 'wikilink1'})
                text_node = wikilink1 and wikilink1.next and wikilink1.next.next or '' # I have no idea what's the propose of this legacy code
                if text_node.strip():
                    rev['sum'] = html.unescape(text_node).strip(u'\u2013 \n')
                    show_edge_case_warning(reason='sum_span not found and text_node found', r_url=r.url, rev_sum=rev['sum'],
                    wikilink1=wikilink1.decode(),
                    next1=wikilink1.next.decode() if wikilink1.next else None,
                    next2=wikilink1.next.next.decode() if (wikilink1.next and wikilink1.next.next) else None)

            # date: optional(str)
            date_span = li.find('span', {'class': 'date'})
            if date_span:
                rev['date'] = date_span.text.strip()
            else:
                rev['date'] = ' '.join(li.text.strip().split(' ')[:2])
                matches = re.findall(
                    r'([0-9./]+ [0-9]{1,2}:[0-9]{1,2})',
                    rev['date'])
                if matches:
                    rev['date'] = matches[0]

            # sizechange: optional(int)
            sizechange_span = li.find('span', {'class': 'sizechange'})

            if sizechange_span:
                sizechange_text = sizechange_span.text.replace('\xC2\xA0', ' ').strip()
                units = ['B', 'KB', 'MB', 'GB']
                positive = '−' not in sizechange_text
                size_change = re.sub(r'[^0-9.]', '', sizechange_text)
                try:
                    size_change = float(size_change)
                except ValueError:
                    size_change = 0.0

                for unit in units[1:]:
                    if unit in sizechange_text:
                        size_change *= 1024
                rev['sizechange'] = positive and int(size_change) or int(-size_change)

            # user: optional(str)
            # legacy
            # if not (select_revs and len(revs) > i and revs[i]['user']):
            user_span = li.find('span', {'class': 'user'})
            if user_span and user_span.text is not None:
                rev['user'] = html.unescape(user_span.text).strip()

            # if select_revs and len(revs) > i:
            #     revs[i].update(rev)
            # else:
            #     revs.append(rev)

            _rev: Revision = {**rev_tmplate,**rev}  # merge dicts # type: ignore
            revs.append(_rev)

            i += 1

        # next page
        first = soup.find_all('input', {'name': 'first', 'value': True})
        continue_index = first and max(map(lambda x: int(x['value']), first))
        cont = soup.find('input', {'class': 'button', 'accesskey': 'n'})
        # time.sleep(1.5)

    # if revs and use_hidden_rev and not select_revs:
    #     soup2 = BeautifulSoup(session.get(url, params={'id': title}).text)
    #     revs[0]['id'] = soup2.find(
    #         'input', {
    #             'type': 'hidden', 'name': 'rev', 'value': True})['value']

    return revs


DATE_FORMATS = ["%Y-%m-%d %H:%M", # <https://www.dokuwiki.org/dokuwiki?do=revisions>
                "%Y-%m-%d", # <http://neff.family.name/unwiki/doku.php>
                "%Y/%m/%d", # <https://tjgrant.com/wiki/news?do=revisions>
                "%Y/%m/%d %H:%M",
                "%Y-%m-%d %H:%M:%S",
                "%Y/%m/%d %H:%M:%S",
                
                "%d.%m.%Y %H:%M",
                "%d/%m/%Y %H:%M", # <https://eolienne.f4jr.org/?do=revisions>
                "%d.%m.%Y %H:%M:%S",
                "%d/%m/%Y %H:%M:%S", # <https://aezoo.compute.dtu.dk/doku.php>

                "%d/%m/%Y alle %H:%M", # <http://didawiki.cli.di.unipi.it/> # 01/03/2007 alle 14:20 (16 anni fa)

                "Le %d/%m/%Y, %H:%M", # <https://doc.ubuntu-fr.org> # Le 23/09/2020, 17:01

                "%H:%M %d/%m/%Y", # <https://lsw.wiki/>
                "%d. %m. %Y (%H:%M)", # <https://www.hks.re/wiki/>
                ]
""" Why there are so many date formats in the world? :( """

def save_page_changes(dumpDir, title: str, revs: List[Revision], child_path, msg_header: str):
    changes_file = dumpDir + '/meta/' + title.replace(':', '/') + '.changes'
    if os.path.exists(changes_file):
        print(msg_header, '    meta change file exists:', changes_file)
        return

    revidOfPage: set[str] = set()
    rows2write = []
    # Loop through revisions in reverse.
    for rev in revs[::-1]:
        print(msg_header, '    meta change saving:', rev)
        summary = rev['sum'] and rev['sum'].strip() or ''
        rev_id = str(0)

        ip = '127.0.0.1'
        user = ''
        minor = 'minor' in rev and rev['minor']

        if 'id' in rev and rev['id']:
            rev_id = rev['id']
        else:
            # Different date formats in different versions of DokuWiki.
            # If no ID was found, make one up based on the date (since rev IDs are Unix times)
            # Maybe this is evil. Not sure.

            print(msg_header, '    One revision of [[%s]] missing rev_id. Using date to rebuild...'
                    % title, end=' ')
            
            assert rev['date'] is not None, 'unknown rev id and unknown rev date: %s' % rev

            for date_format in DATE_FORMATS:
                try:
                    date = datetime.strptime(
                        # remove " (x days ago)" in f"{date_format} (x days ago)" if date_format not contain '('
                        rev['date'].split('(')[0].strip(),
                        date_format
                    ) if '(' not in date_format else datetime.strptime(
                        # date_format contain '('
                        rev['date'].strip(),
                        date_format
                    )
                    rev_id = str(int(time.mktime(date.utctimetuple())))
                    break
                except Exception:
                    rev_id = None
                    
            assert rev_id is not None, 'Cannot parse date: %s' % rev['date']
            assert isinstance(rev_id, str), 'rev_id must be str, not %s' % type(rev_id)

            # if rev_id is not unique, plus 1 to it until it is.
            while rev_id in revidOfPage:
                rev_id = str(int(rev_id) + 1)
            print(msg_header, 'rev_id is now %s' % rev_id)

        revidOfPage.add(rev_id)

        rev['user'] = rev['user'] if rev['user'] is not None else 'unknown'
        try:
            ip_parsed = ip_address(rev['user'])
            assert isinstance(ip_parsed, (IPv4Address, IPv6Address))
            ip = rev['user']
        except ValueError:
            user = rev['user']

        sizechange = rev['sizechange'] or ''

        extra = ''  # TODO: use this
        # max 255 chars(utf-8) for summary. (dokuwiki limitation)
        summary = summary[:255]
        row = '\t'.join([rev_id, ip, 'e' if minor else 'E',
                        title, user, summary, extra, str(sizechange)])
        row = row.replace('\n', ' ')
        row = row.replace('\r', ' ')
        rows2write.append(row)


    smkdirs(dumpDir, '/meta/' + child_path)
    with uopen(changes_file, 'w') as f:
        f.write('\n'.join(rows2write)+'\n')



import argparse
import hashlib
import shutil
import time
import os
import subprocess
from dataclasses import dataclass
from typing import Dict, List, Tuple, Optional

from internetarchive import get_item, Item
import requests

from dokuWikiDumper.utils.util import url2prefix
from dokuWikiDumper.dump.info.info import get_info
from dokuWikiDumper.dump.info.info import INFO_WIKI_NAME, INFO_RAW_TITLE, INFO_DOKU_URL, INFO_LANG
from dokuWikiDumper.utils.config import get_config

from .__version__ import UPLOADER_VERSION


# Constants
DEFAULT_COLLECTION = 'opensource'
USER_AGENT = f'dokuWikiUploader/{UPLOADER_VERSION}'
UPLOADED_MARK = 'uploaded_to_IA.mark'
BUFFER_SIZE = 65536

# Directory configurations
DIRS_TO_7Z = ["attic", "html", "media", "pages", "pdf", "meta"]
MARK_FILES = {
    "attic": "content_dumped.mark",
    "pages": "content_dumped.mark", 
    "html": "html_dumped.mark",
    "media": "media_dumped.mark",
    "pdf": "pdf_dumped.mark",
    "dumpMeta": "dumpMeta/",
    "meta": "meta/"
}

# Compression levels
MEDIA_COMPRESSION_LEVEL = 1
DEFAULT_COMPRESSION_LEVEL = 5
NO_COMPRESSION_LEVEL = 0


@dataclass
class UploadConfig:
    """Configuration for upload process."""
    dump_dir: str
    path7z: str
    access_key: str
    secret_key: str
    collection: str
    pack_dumpMeta_dir: bool
    level0_no_compress: List[str]
    delete_after_upload: bool = False


@dataclass
class WikiMetadata:
    """Wiki metadata container."""
    name: str
    url: Optional[str]
    language: Optional[str]
    raw_title: Optional[str]


class IAUploader:
    """Internet Archive uploader for DokuWiki dumps."""
    
    def __init__(self, config: UploadConfig):
        self.config = config
        self.headers = {"User-Agent": USER_AGENT}
        
    def upload_dump(self) -> None:
        """Main upload process."""
        if self._is_already_uploaded():
            print("This dump has already been uploaded.")
            print(f"If you want to upload it again, please remove the file '{UPLOADED_MARK}'.")
            return
            
        try:
            # Prepare metadata and files
            wiki_meta = self._extract_wiki_metadata()
            identifier = self._generate_identifier()
            files_to_upload = self._prepare_files(identifier)
            
            # Upload to Internet Archive
            item_metadata = self._create_item_metadata(wiki_meta)
            self._upload_to_ia(identifier, files_to_upload, item_metadata)
            
            # Update metadata with URLs
            self._update_item_metadata(identifier, wiki_meta)
            
            # Mark as uploaded
            self._mark_as_uploaded()
            
            # Cleanup if requested
            if self.config.delete_after_upload:
                self._cleanup_dump_dir()
                
            print(f"\n\n--Done--\nYou can find it in https://archive.org/details/wiki-{identifier}")
            
        except Exception as e:
            print(f"Upload failed: {e}")
            raise
    
    def _is_already_uploaded(self) -> bool:
        """Check if dump is already uploaded."""
        return os.path.exists(os.path.join(self.config.dump_dir, UPLOADED_MARK))
    
    def _extract_wiki_metadata(self) -> WikiMetadata:
        """Extract wiki metadata from dump directory."""
        info = get_info(self.config.dump_dir)
        config = get_config(self.config.dump_dir)
        
        # Determine wiki name
        wiki_name = (info.get(INFO_WIKI_NAME) or 
                     info.get(INFO_RAW_TITLE))
        
        # Fallback to URL-based name if no name found
        if not wiki_name:
            if doku_url := info.get(INFO_DOKU_URL):
                wiki_name = url2prefix(doku_url, ascii_slugify=False)
        
        # Final fallback
        wiki_name = wiki_name or 'Unknown'
        
        # Handle special case for 'start' pages:
        # some sites use 'wikiname [pagetitle]' as the title instead of 'pagetitle [wikiname]'.
        # so, fallback to INFO_RAW_TITLE
        if wiki_name == 'start' and info.get(INFO_RAW_TITLE):
            raw_title = info.get(INFO_RAW_TITLE)
            if raw_title:
                wiki_name = raw_title.replace('[start]', '')
        
        # Clean up wiki name
        wiki_name = wiki_name.replace('\r', '').replace('\n', '').strip()
        
        return WikiMetadata(
            name=wiki_name,
            url=info.get(INFO_DOKU_URL),
            language=info.get(INFO_LANG) or config.get('lang'),
            raw_title=info.get(INFO_RAW_TITLE)
        )
    
    def _generate_identifier(self) -> str:
        """Generate identifier from dump directory name."""
        dump_dir_ab = os.path.abspath(self.config.dump_dir)
        dump_dir_basename = os.path.basename(dump_dir_ab)
        
        # Validate dump directory name format
        parts = dump_dir_basename.split('-') # Punycoded domain may contain multiple `-`
        if len(parts) < 2:
            raise ValueError(f'Invalid dump directory name: {dump_dir_basename}')
        
        try:
            timestamp = int(parts[-1])
            if timestamp < 20230200:
                raise ValueError(
                    f'Invalid dump directory name: {dump_dir_basename}, '
                    'created before dokuWikiDumper was born!?'
                )
        except ValueError as e:
            if "created before" not in str(e):
                raise ValueError(f'Invalid dump directory name: {dump_dir_basename}')
            raise
        
        return dump_dir_basename
    
    def _prepare_files(self, identifier: str) -> Dict[str, str]:
        """Prepare files for upload."""
        filedict = {}
        
        # Handle dumpMeta directory
        dirs_to_pack = DIRS_TO_7Z.copy()
        if self.config.pack_dumpMeta_dir:
            if "dumpMeta" not in dirs_to_pack:
                dirs_to_pack.append("dumpMeta")
        else:
            self._add_dumpmeta_files(identifier, filedict)
        
        # Compress and add directories
        for dir_name in dirs_to_pack:
            dir_path = os.path.join(self.config.dump_dir, dir_name)
            if os.path.isdir(dir_path):
                self._validate_directory_completion(dir_name)
                compressed_file = self._compress_directory(dir_path, dir_name)
                filedict[f"{identifier}-{dir_name}.7z"] = compressed_file
        
        return filedict
    
    def _add_dumpmeta_files(self, identifier: str, filedict: Dict[str, str]) -> None:
        """Add individual dumpMeta files to upload list."""
        dumpmeta_dir = os.path.join(self.config.dump_dir, "dumpMeta")
        if os.path.exists(dumpmeta_dir):
            for item in os.listdir(dumpmeta_dir):
                item_path = os.path.join(dumpmeta_dir, item)
                remote_name = f"{identifier}-dumpMeta/{item}"
                filedict[remote_name] = item_path
    
    def _validate_directory_completion(self, dir_name: str) -> None:
        """Validate that directory dumping is complete."""
        mark_file = MARK_FILES.get(dir_name)
        if mark_file and not mark_file.endswith('/'):
            mark_path = os.path.join(self.config.dump_dir, mark_file)
            if not os.path.exists(mark_path):
                raise Exception(
                    f"Directory {dir_name} is not finished. "
                    f"Please run dokuWikiDumper again. ({mark_file} not found)"
                )
    
    def _compress_directory(self, dir_path: str, dir_name: str) -> str:
        """Compress directory to 7z format."""
        print(f"Compressing {dir_path}...")
        
        # Determine compression level
        if dir_name in self.config.level0_no_compress:
            level = NO_COMPRESSION_LEVEL
            print(f"Packing {dir_name} with level 0 compression...")
        elif dir_name in ["media", "pdf"]:
            level = MEDIA_COMPRESSION_LEVEL
        else:
            level = DEFAULT_COMPRESSION_LEVEL
        
        return self._compress_with_7z(dir_path, level)
    
    def _compress_with_7z(self, dir_path: str, level: int) -> str:
        """Compress directory using 7z."""
        dir_path = os.path.abspath(dir_path)
        output_file = f"{dir_path}.7z"
        temp_file = f"{output_file}.tmp"
        
        if os.path.exists(output_file):
            print(f"File {output_file} already exists. Skip compressing.")
            return output_file
        
        # Build command
        if level == NO_COMPRESSION_LEVEL:
            cmd = [
                self.config.path7z, "a", "-t7z", f"-mx={level}",
                "-scsUTF-8", "-ms=off", temp_file, dir_path
            ]
        else:
            cmd = [
                self.config.path7z, "a", "-t7z", "-m0=lzma2", f"-mx={level}",
                "-scsUTF-8", "-md=64m", "-ms=off", temp_file, dir_path
            ]
        
        subprocess.run(cmd, check=True)
        os.rename(temp_file, output_file)
        return output_file
    
    def _create_item_metadata(self, wiki_meta: WikiMetadata) -> Dict[str, str]:
        """Create initial item metadata."""
        config = get_config(self.config.dump_dir)
        
        keywords = ["wiki", "wikiteam", "DokuWiki", "dokuWikiDumper", "wikidump"]
        if wiki_meta.name and wiki_meta.name not in keywords:
            keywords.append(wiki_meta.name)
        if wiki_meta.url:
            keywords.append(url2prefix(wiki_meta.url, ascii_slugify=False))
        
        description = (
            f"DokuWiki: {wiki_meta.name}\n<br>\n<br>\n"
            f"Dumped with DokuWiki-Dumper v{config.get('dokuWikiDumper_version')}, "
            f"and uploaded with dokuWikiUploader v{UPLOADER_VERSION}."
        )
        
        metadata = {
            "mediatype": "web",
            "collection": self.config.collection,
            "title": f"Wiki - {wiki_meta.name}",
            "description": description,
            "last-updated-date": time.strftime("%Y-%m-%d", time.gmtime()),
            "subject": "; ".join(keywords[:5]),  # Initial keywords only
            "upload-state": "uploading",
        }
        
        if wiki_meta.language:
            metadata["language"] = wiki_meta.language
            
        return metadata
    
    def _upload_to_ia(self, identifier: str, files: Dict[str, str], metadata: Dict[str, str]) -> None:
        """Upload files to Internet Archive."""
        remote_identifier = f"wiki-{identifier}"
        
        print(f"Identifier (Local): {identifier}")
        print(f"Identifier (Remote): {remote_identifier}")
        
        # Check for existing files
        item = get_item(remote_identifier)
        files_to_upload = self._filter_existing_files(item, files)
        
        print(f"Uploading {len(files_to_upload)} files...")
        print(metadata)
        
        if files_to_upload:
            item.upload(
                files=files_to_upload,
                metadata=metadata,
                access_key=self.config.access_key,
                secret_key=self.config.secret_key,
                verbose=True,
                queue_derive=False,
            )
        
        print(f"Uploading {len(files_to_upload)} files: Done.\n")
        
        # Wait for item to be created
        self._wait_for_item_creation(remote_identifier)
    
    def _filter_existing_files(self, item: Item, files: Dict[str, str]) -> Dict[str, str]:
        """Filter out files that already exist in the item."""
        existing_files = {f["name"] for f in item.files}
        filtered_files = {}
        
        for remote_name, local_path in files.items():
            if remote_name in existing_files:
                print(f"File {remote_name} already exists in item.")
            else:
                filtered_files[remote_name] = local_path
                
        return filtered_files
    
    def _wait_for_item_creation(self, identifier: str, max_tries: int = 30) -> None:
        """Wait for item to be created on Internet Archive."""
        tries = max_tries
        item = get_item(identifier)
        
        while not item.exists and tries > 0:
            print(f"Waiting for item to be created ({tries})...", end='\r')
            time.sleep(30)
            item = get_item(identifier)
            tries -= 1
    
    def _update_item_metadata(self, identifier: str, wiki_meta: WikiMetadata) -> None:
        """Update item metadata with URLs and final information."""
        remote_identifier = f"wiki-{identifier}"
        item = get_item(remote_identifier)
        
        print("Updating description...")
        
        updates = {}
        
        # Update description with URL
        if (wiki_meta.url and 
            (wiki_meta.url not in item.metadata.get("description", "") or
             'https://github.com/saveweb/dokuwiki-dumper' not in item.metadata.get("description", ""))):
            
            description_with_url = (
                f'DokuWiki: <a href="{wiki_meta.url}" rel="nofollow">{wiki_meta.name}</a>\n'
                '<br>\n<br>\n'
                f'Dumped with <a href="https://github.com/saveweb/dokuwiki-dumper" rel="nofollow">'
                f'DokuWiki-Dumper</a> v{get_config(self.config.dump_dir).get("dokuWikiDumper_version")}, '
                f'and uploaded with dokuWikiUploader v{UPLOADER_VERSION}.'
            )
            updates["description"] = description_with_url
        
        # Update other metadata fields
        current_date = time.strftime("%Y-%m-%d", time.gmtime())
        if item.metadata.get("last-updated-date") != current_date:
            updates["last-updated-date"] = current_date
        
        # Update subject with length limits
        subject = self._create_subject_string(wiki_meta)
        if item.metadata.get("subject") != subject:
            updates["subject"] = subject
        
        if wiki_meta.url and item.metadata.get("originalurl") != wiki_meta.url:
            updates["originalurl"] = wiki_meta.url
        
        if item.metadata.get("upload-state") != "uploaded":
            updates["upload-state"] = "uploaded"
        
        # Apply updates
        if updates:
            response = item.modify_metadata(
                metadata=updates,
                access_key=self.config.access_key,
                secret_key=self.config.secret_key
            )
            
            if isinstance(response, requests.Response):
                print(response.text)
                response.raise_for_status()
                print("Updating description: Done.")
            else:
                print("Unexpected response type during metadata update")
        else:
            print("Updating description: No need to update.")
    
    def _create_subject_string(self, wiki_meta: WikiMetadata) -> str:
        """Create subject string respecting IA's 255 byte limit."""
        base_keywords = ["wiki", "wikiteam", "DokuWiki", "dokuWikiDumper", "wikidump"]
        all_keywords = base_keywords.copy()
        
        if wiki_meta.name and wiki_meta.name not in all_keywords:
            all_keywords.append(wiki_meta.name)
        if wiki_meta.url:
            all_keywords.append(url2prefix(wiki_meta.url, ascii_slugify=False))
        
        # Try full keywords first
        full_subject = "; ".join(all_keywords)
        if len(full_subject.encode("utf-8")) <= 255:
            return full_subject
        
        # Try without wiki name
        without_name = base_keywords + ([url2prefix(wiki_meta.url, ascii_slugify=False)] if wiki_meta.url else [])
        subject_without_name = "; ".join(without_name)
        if len(subject_without_name.encode("utf-8")) <= 255:
            return subject_without_name
        
        # Fallback to base keywords only
        return "; ".join(base_keywords)
    
    def _mark_as_uploaded(self) -> None:
        """Mark dump as uploaded."""
        timestamp = time.strftime("%Y-%m-%d %H:%M:%S")
        mark_content = f"Uploaded to Internet Archive with dokuWikiUploader v{UPLOADER_VERSION} on {timestamp}"
        
        mark_path = os.path.join(self.config.dump_dir, UPLOADED_MARK)
        with open(mark_path, "w", encoding='UTF-8') as f:
            f.write(mark_content)
    
    def _cleanup_dump_dir(self) -> None:
        """Delete dump directory after upload."""
        print("Deleting the dump dir in 5 seconds...(Ctrl+C to cancel)", end="")
        for _ in range(5):
            print(".", end="", flush=True)
            time.sleep(1)
        shutil.rmtree(self.config.dump_dir)
        print("Deleted!")


# Utility functions
def file_sha1(path: str) -> str:
    """Calculate SHA1 hash of a file."""
    buffer = bytearray(BUFFER_SIZE)
    view = memoryview(buffer)
    digest = hashlib.sha1()

    with open(path, mode="rb") as f:
        while True:
            n = f.readinto(buffer)
            if not n:
                break
            digest.update(view[:n])

    return digest.hexdigest()


def read_ia_keys(keysfile: str) -> Tuple[str, str]:
    """Read Internet Archive keys from file.
    
    Returns:
        Tuple of (access_key, secret_key)
    """
    keysfile = os.path.expanduser(keysfile)
    
    
    with open(keysfile, 'r', encoding='utf-8') as f:
        lines = f.readlines()
    
    if len(lines) < 2:
        raise ValueError("Keys file must contain at least 2 lines")
    
    access_key = lines[0].strip()
    secret_key = lines[1].strip()
        
    return access_key, secret_key


def create_upload_config(args: argparse.Namespace) -> UploadConfig:
    """Create upload configuration from command line arguments."""
    access_key, secret_key = read_ia_keys(args.keysfile)
    
    return UploadConfig(
        dump_dir=args.dump_dir,
        path7z=args.path7z,
        access_key=access_key,
        secret_key=secret_key,
        collection=args.collection,
        pack_dumpMeta_dir=args.pack_dumpMeta,
        level0_no_compress=args.level0_no_compress or [],
        delete_after_upload=args.delete
    )


def create_argument_parser() -> argparse.ArgumentParser:
    """Create and configure argument parser."""
    parser = argparse.ArgumentParser(
        prog="dokuWikiUploader",
        description=f"Upload a DokuWiki dump to Internet Archive. (Version: {UPLOADER_VERSION})."
    )
    
    parser.add_argument(
        "-kf", "--keysfile", 
        default="~/.doku_uploader_ia_keys",
        help="Path to the IA S3 keys file. (first line: access key, second line: secret key) "
             "[default: ~/.doku_uploader_ia_keys]"
    )
    
    parser.add_argument(
        "-p7z", "--path7z", 
        default="7z",
        help="Path to 7z binary. [default: 7z]"
    )
    
    parser.add_argument(
        "-c", "--collection", 
        default=DEFAULT_COLLECTION,
        help="Collection to upload to. ('test_collection' for testing (auto-delete after 30 days) "
             "[default: opensource]"
    )
    
    parser.add_argument(
        "-p", "--pack-dumpMeta", 
        action="store_true",
        help="Pack the dumpMeta/ directory into a 7z file, then upload it. "
             "instead of uploading all files in dumpMeta/ directory individually. "
             "[default: False]"
    )
    
    parser.add_argument(
        '-n', '--level0-no-compress', 
        default=[], 
        dest='level0_no_compress',
        choices=['media', 'pdf'], 
        nargs='?', 
        action='append',
        help='Pack specified dir(s) into 7z file(s) without any compression. (level 0, copy mode)'
    )
    
    parser.add_argument(
        '-d', '--delete', 
        action='store_true', 
        dest='delete',
        help='Delete the dump dir after uploading. [default: False]'
    )
    
    parser.add_argument(
        "dump_dir", 
        help="Path to the wiki dump directory."
    )
    
    return parser


def main(params: Optional[List[str]] = None) -> int:
    """Main entry point."""
    parser = create_argument_parser()
    args = parser.parse_args(params)
    
    try:
        config = create_upload_config(args)
        uploader = IAUploader(config)
        uploader.upload_dump()
        return 0
        
    except KeyboardInterrupt:
        print("\nUpload cancelled by user.")
        return 1
    except Exception as e:
        print(f"Upload failed: {e}")
        return 1


if __name__ == "__main__":
    exit(main())



from dokuWikiDumper.exceptions import VersionOutdatedError
from dokuWikiDumper.version import get_latest_version, get_version


def dokuWikiDumper_outdated_check():
    latest_version = get_latest_version()
    if latest_version is None:
        return
    if latest_version != get_version():
        print('=' * 47)
        print(f'Warning: You are using an outdated version of dokuWikiDumper ({get_version()}).')
        print(f'         The latest version is {latest_version}.')
        print('         You can update dokuWikiDumper with "pip3 install --upgrade dokuWikiDumper".')
        print('=' * 47, end='\n\n')
        raise VersionOutdatedError(version=get_version())
    print('You are using the latest version of dokuwikidumper.')



import copy
from dataclasses import dataclass
import os
import queue
import threading
import concurrent.futures

import requests
from dokuWikiDumper.dump.content.revisions import get_revisions
from dokuWikiDumper.dump.content.titles import load_get_save_titles
from dokuWikiDumper.utils.util import smkdirs
from dokuWikiDumper.utils.util import print_with_lock as print

from dokuWikiDumper.exceptions import DispositionHeaderMissingError

PDF_DIR = 'pdf/'
PDF_PAGR_DIR = PDF_DIR + 'pages/'
PDF_OLDPAGE_DIR = PDF_DIR + 'attic/'

exit_event = threading.Event()


@dataclass
class DumpPDFParams:
    dump_dir: str
    title: str
    title_index: int
    doku_url: str
    session: requests.Session
    current_only: bool


def dump_PDF(doku_url, dump_dir,
                  session: requests.Session, threads: int = 1,
                  ignore_errors: bool = False, current_only: bool = False):
    titles = load_get_save_titles(dump_dir=dump_dir, url=doku_url, session=session)
    
    if not len(titles):
        print('Empty wiki')
        return False
    
    task_templ = DumpPDFParams(dump_dir=dump_dir, doku_url=doku_url, session=session, current_only=current_only,
                              title_index=-999, title="dokuwikidumper_placehold")

    tasks_queue: queue.Queue[DumpPDFParams] = queue.Queue(maxsize=threads)

    def task_generator():
        for index, title in enumerate(titles):
            task = copy.copy(task_templ)
            task.title_index = index
            task.title = title
            tasks_queue.put(task)
            print('PDF: (%d/%d): [[%s]] ...' % (index+1, len(titles), title))

            if exit_event.is_set():
                print('task generator exit (exit event set)')
                return

        tasks_queue.join()
        print('All tasks done, terminating workers...')
        exit_event.set()

    tg_thread = threading.Thread(target=task_generator, name='task-generator')
    tg_thread.daemon = True
    tg_thread.start()

    def _dump_pdf_action(task: DumpPDFParams, ignore_errors: bool):
        try:
            dump_pdf_page(task)
        except Exception as e:
            if not ignore_errors:
                raise e
            else:
                print('[',task.title_index+1,'] Error in sub thread: (', e, ') ignored')

    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        futures = set()
        while not exit_event.is_set():
            try:
                task = tasks_queue.get(timeout=1)
            except queue.Empty:
                continue
            f = executor.submit(_dump_pdf_action, task, ignore_errors)
            f.add_done_callback(lambda f: tasks_queue.task_done())
            futures.add(f)

            if len(futures) >= threads:
                _done, futures = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
                for f in _done:
                    f.result()

        while futures:
            _done, futures = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
            for f in _done:
                f.result()

    tg_thread.join()


def dump_pdf_page(task: DumpPDFParams):
    msg_header = '['+str(task.title_index + 1)+']: '

    file = task.dump_dir + '/' + PDF_PAGR_DIR + task.title.replace(':', '/') + '.pdf'
    local_size = -1
    if os.path.isfile(file):
        local_size = os.path.getsize(file)
    with task.session.get(task.doku_url, params={'do': 'export_pdf', 'id': task.title}, stream=True) as r:
        r.raise_for_status()
        if 'Content-Disposition' not in r.headers:
            raise DispositionHeaderMissingError(r)
        remote_size = int(r.headers.get('Content-Length', -2))

        if local_size == remote_size:
            print(msg_header, '[[%s]]' % task.title, 'already exists')
        else:
            child_path = task.title.replace(':', '/')
            child_dir = os.path.dirname(child_path)
            smkdirs(task.dump_dir, PDF_PAGR_DIR, child_dir)
            with open(file, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
            print(msg_header, '[[%s]]' % task.title, 'saved')

    if task.current_only:
        return True

    revs = get_revisions(doku_url=task.doku_url, session=task.session, title=task.title, msg_header=msg_header)

    for rev in revs[1:]:
        if 'id' in rev and rev['id']:
            try:
                r = task.session.get(task.doku_url, params={'do': 'export_pdf', 'id': task.title, 'rev': rev['id']})
                r.raise_for_status()
                content = r.content
                smkdirs(task.dump_dir, PDF_OLDPAGE_DIR, child_dir)
                old_pdf_path = task.dump_dir + '/' + PDF_OLDPAGE_DIR + child_path + '.' + rev['id'] + '.pdf'

                with open(old_pdf_path, 'bw') as f:
                    f.write(content)
                print(msg_header, '    Revision %s of [[%s]] saved.' % (rev['id'], task.title))
            except requests.HTTPError as e:
                print(msg_header, '    Revision %s of [[%s]] failed: %s' % (rev['id'], task.title, e))


from dokuWikiDumper.utils.util import check_int

def test_check_int():
    assert 1 == check_int(1)
    assert check_int("a") is None
    assert "1" == check_int("1")
    assert 1.1 == check_int(1.1)
    assert None is check_int(None)
    



import copy
from dataclasses import dataclass
import os
import queue
import threading
import concurrent.futures
import requests

from dokuWikiDumper.dump.content.revisions import get_revisions, save_page_changes
from dokuWikiDumper.dump.content.titles import load_get_save_titles

from dokuWikiDumper.utils.util import smkdirs, uopen
from dokuWikiDumper.utils.util import print_with_lock as print
from dokuWikiDumper.utils.config import runtime_config

HTML_DIR = 'html/'
HTML_PAGR_DIR = HTML_DIR + 'pages/'
HTML_OLDPAGE_DIR = HTML_DIR + 'attic/'

exit_event = threading.Event()

@dataclass
class DumpHTMLParams:
    dump_dir: str
    title_index: int
    title: str
    doku_url: str
    session: requests.Session
    current_only: bool


def dump_HTML(doku_url, dump_dir,
                session: requests.Session, threads: int = 1,
                ignore_errors: bool = False, current_only: bool = False):
    smkdirs(dump_dir, HTML_PAGR_DIR)

    titles = load_get_save_titles(dump_dir=dump_dir, url=doku_url, session=session)
    
    if not len(titles):
        print('Empty wiki')
        return False

    task_templ = DumpHTMLParams(dump_dir=dump_dir, title_index=-999, title='', doku_url=doku_url, session=session, current_only=current_only)

    tasks_queue: queue.Queue[DumpHTMLParams] = queue.Queue(maxsize=threads)

    def task_generator():
        for index, title in enumerate(titles):
            task = copy.copy(task_templ)
            task.title_index = index
            task.title = title
            tasks_queue.put(task)
            print('HTML: (%d/%d): [[%s]] ...' % (index+1, len(titles), title))

            if exit_event.is_set():
                print('task generator exit (exit event set)')
                return

        tasks_queue.join()
        print('All tasks done, terminating workers...')
        exit_event.set()

    tg_thread = threading.Thread(target=task_generator, name='task-generator')
    tg_thread.daemon = True
    tg_thread.start()

    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        futures = set()
        while not exit_event.is_set():
            try:
                task = tasks_queue.get(timeout=1)
            except queue.Empty:
                continue
            f = executor.submit(_dump_html_action, task, ignore_errors)
            f.add_done_callback(lambda f: tasks_queue.task_done())
            futures.add(f)

            if len(futures) >= threads:
                _done, futures = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
                for f in _done:
                    f.result()

        while futures:
            _done, futures = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
            for f in _done:
                f.result()

    tg_thread.join()


def _dump_html_action(task: DumpHTMLParams, ignore_errors: bool):
    try:
        dump_html_page(task)
    except Exception as e:
        if not ignore_errors:
            raise e
        else:
            print('[',task.title_index+1,'] Error in sub thread: (', e, ') ignored')


def dump_html_page(task: DumpHTMLParams):
    r = task.session.get(task.doku_url, params={'do': runtime_config.export_xhtml_action, 'id': task.title})
    # export_html is a alias of export_xhtml, but not exist in older versions of dokuwiki
    r.raise_for_status()
    if r.text is None or r.text == '':
        raise Exception('Empty response (r.text)')

    msg_header = '['+str(task.title_index + 1)+']: '

    title2path = task.title.replace(':', '/')
    child_path = os.path.dirname(title2path)
    html_path = task.dump_dir + '/' + HTML_PAGR_DIR + title2path + '.html'
    smkdirs(task.dump_dir, HTML_PAGR_DIR, child_path)
    with uopen(html_path, 'w') as f:
        f.write(r.text)
        print(msg_header, '[[%s]]' % task.title, 'saved')

    if task.current_only:
        return True

    revs = get_revisions(doku_url=task.doku_url, session=task.session, title=task.title, msg_header=msg_header)

    for rev in revs[1:]:
        if 'id' in rev and rev['id']:
            try:
                r = task.session.get(task.doku_url, params={'do': runtime_config.export_xhtml_action, 'id': task.title, 'rev': rev['id']})
                r.raise_for_status()
                if r.text is None or r.text == '':
                    raise Exception('Empty response (r.text)')
                smkdirs(task.dump_dir, HTML_OLDPAGE_DIR, child_path)
                old_html_path = task.dump_dir + '/' + HTML_OLDPAGE_DIR + title2path + '.' + rev['id'] + '.html'

                with uopen(old_html_path, 'w') as f:
                    f.write(r.text)
                print(msg_header, '    Revision %s of [[%s]] saved.' % (rev['id'], task.title))
            except requests.HTTPError as e:
                print(msg_header, '    Revision %s of [[%s]] failed: %s' % (rev['id'], task.title, e))
        else:
            print(msg_header, '    Revision %s of [[%s]] failed: %s' % (rev['id'], task.title, 'Rev id not found (please check ?do=revisions of this page)'))

    save_page_changes(dumpDir=task.dump_dir, child_path=child_path, title=task.title, 
                       revs=revs, msg_header=msg_header)


# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import os
import sys
import time

import requests

from dokuWikiDumper.utils.dump_lock import DumpLock
from dokuWikiDumper.utils.ia_checker import any_recent_ia_item_exists
from dokuWikiDumper.utils.util import print_with_lock as print

from dokuWikiDumper.version import get_version
from dokuWikiDumper.version_check import dokuWikiDumper_outdated_check
from dokuWikiDumper.dump.content.content import dump_content
from dokuWikiDumper.dump.html.html import dump_HTML
from dokuWikiDumper.dump.info.info import update_info
from dokuWikiDumper.dump.media.media import dump_media
from dokuWikiDumper.dump.pdf.pdf import dump_PDF
from dokuWikiDumper.utils.config import update_config, runtime_config
from dokuWikiDumper.utils.patch import SessionMonkeyPatch
from dokuWikiDumper.utils.session import create_session, load_cookies, login_dokuwiki
from dokuWikiDumper.utils.util import avoidSites, build_base_url, get_doku_url, smkdirs, standardize_url, url2prefix

DEFAULT_THREADS = -1 # magic number, -1 means use 1 thread.

def getArgumentParser():
    parser = argparse.ArgumentParser(description='dokuWikiDumper Version: '+ get_version())
    parser.add_argument('url', help='URL of the dokuWiki (provide the doku.php URL)', type=str)

    group_download = parser.add_argument_group("Data to download", "What info download from the wiki")

    group_download.add_argument('--content', action='store_true', help='Dump content')
    group_download.add_argument('--media', action='store_true', help='Dump media')
    group_download.add_argument('--html', action='store_true', help='Dump HTML')
    group_download.add_argument('--pdf', action='store_true',
                        help='Dump PDF [default: false] (Only available on some wikis with the PDF export plugin)'+
                        ' (Only dumps the latest PDF revision)')

    parser.add_argument('--current-only', dest='current_only', action='store_true',
                        help='Dump latest revision, no history [default: false]')
    # TODO: add back
    # parser.add_argument(
    #     '--skip-to', help='!DEV! Skip to title number [default: 0]', type=int, default=0)
    parser.add_argument(
        '--path', help='Specify dump directory [default: <site>-<date>]', type=str, default='')
    parser.add_argument(
        '--no-resume', help='Do not resume a previous dump [default: resume]', action='store_true')
    parser.add_argument(
        '--threads', help='Number of sub threads to use [default: 1], not recommended to set > 5', type=int, default=DEFAULT_THREADS)
    parser.add_argument(
        '--i-love-retro',  action='store_true', dest='user_love_retro',
        help='Do not check the latest version of dokuWikiDumper (from pypi.org) before running [default: False]')

    parser.add_argument('--insecure', action='store_true',
                        help='Disable SSL certificate verification')
    parser.add_argument('--ignore-errors', action='store_true',
                        help='!DANGEROUS! ignore errors in the sub threads. This may cause incomplete dumps.')
    parser.add_argument('--ignore-action-disabled-edit', action='store_true',
                        help='Some sites disable edit action for anonymous users and some core pages. '+
                        'This option will ignore this error and textarea not found error.'+
                        'But you may only get a partial dump. '+
                        '(only works with --content)', dest='ignore_action_disabled_edit')
    parser.add_argument('--trim-php-warnings', action='store_true', dest='trim_php_warnings',
                        help='Trim PHP warnings from requests.Response.text')

    parser.add_argument('--export-xhtml-action', type=str, choices=['export_html', 'export_xhtml'], dest='export_xhtml_action',
                        help='HTML export action [default: export_xhtml]', default='export_xhtml')

    parser.add_argument('--delay', help='Delay between requests [default: 0.0]', type=float, default=0.0)
    parser.add_argument('--retry', help='Maximum number of retries [default: 5]', type=int, default=5)
    parser.add_argument('--hard-retry', type=int, default=3, dest='hard_retry',
                        help='Maximum number of retries for hard errors [default: 3]')

    parser.add_argument('--parser', help='HTML parser [default: lxml]', type=str, default='lxml')

    parser.add_argument('--username', help='login: username')
    parser.add_argument('--password', help='login: password')
    parser.add_argument('--verbose', action='store_true', help='Verbose output')
    parser.add_argument('--cookies', help='cookies file')
    parser.add_argument('--auto', action='store_true', 
                        help='dump: content+media+html, threads=3, ignore-action-disable-edit. (threads is overridable)')
    parser.add_argument('-u', '--upload', action='store_true', 
                        help='Upload wikidump to Internet Archive after successfully dumped'
                        ' (only works with --auto)')
    parser.add_argument("-g", "--uploader-arg", dest="uploader_args", action='append', default=[],
                        help="Arguments for uploader.")
    parser.add_argument('--force', action='store_true', help='To dump even if a recent dump exists on IA')
    parser.add_argument('--user-agent', dest="user_agent", type=str,
                        default='dokuWikiDumper/' + get_version() + ' (https://github.com/saveweb/dokuwiki-dumper)',
                        help=argparse.SUPPRESS)

    return parser


def checkArgs(args):
    if not args.content and not args.media and not args.html and not args.pdf:
        print('Nothing to do. Use --content and/or --media and/or --html and/or --pdf to specify what to dump.')
        return False
    if not args.url:
        print('No URL specified.')
        return False
    if args.threads < 1:
        print('Number of threads must be >= 1.')
        return False
    if args.threads > 5:
        print('Warning: threads > 5 , will bring a lot of pressure to the server.')
        print('Original site may deny your request, even ban our UA.')
        time.sleep(3)
        input('Press Enter to continue...')
    if args.ignore_errors:
        print('Warning: You have chosen to ignore errors in the sub threads. This may cause incomplete dumps.')
        time.sleep(3)
        input('Press Enter to continue...')
    if args.username and not args.password:
        print('Warning: You have specified a username but no password.')
        return False
    if args.ignore_action_disabled_edit and not args.content:
        print('Warning: You have specified --ignore-action-disabled-edit, but you have not specified --content.')
        return False
    if args.delay < 0:
        print('Delay must be >= 0.')
        return False
    if args.delay > 0.0 and args.threads > 1:
        print(f"Warning: You have specified a delay and more than one thread ({args.threads}).")
        print("!!! Delay will be applied to each thread separately !!!")
        time.sleep(3)
        input('Press Enter to continue...')
    if args.retry < 0:
        print('Retry must be >= 0.')
        return False
    if args.upload and not args.auto:
        print('Warning: You have specified --upload, but you have not specified --auto.')
        return False
    if args.parser:
        from bs4 import BeautifulSoup, FeatureNotFound
        try:
            BeautifulSoup("", args.parser)
            runtime_config.html_parser = args.parser
        except FeatureNotFound:
            print("Parser %s not found. Please install first following "
                  "https://www.crummy.com/software/BeautifulSoup/bs4/doc/#installing-a-parser"%(args.parser))
            return False
    if args.export_xhtml_action:
        runtime_config.export_xhtml_action = args.export_xhtml_action
    return True


def getParameters():
    parser = getArgumentParser()
    args = parser.parse_args()

    if args.auto:
        args.content = True
        args.media = True
        args.html = True
        args.threads = args.threads if args.threads >= 1 else 3
        args.ignore_action_disabled_edit = True
    else:
        # reset magic number
        args.threads = args.threads if args.threads >= 1 else 1
    print('Threads:', args.threads)
    if not checkArgs(args):
        parser.print_help()
        sys.exit(1)

    return args


def dump():
    args = getParameters()
    if not args.user_love_retro:
        dokuWikiDumper_outdated_check()
    url_input = args.url

    session = create_session(retries=args.retry, user_agent=args.user_agent)

    if args.verbose:
        def print_request(r: requests.Response, *args, **kwargs):
            # TODO: use logging
            # print("H:", r.request.headers)
            for _r in r.history:
                print("Resp (history): ", _r.request.method, _r.status_code, _r.reason, _r.url)
            print(f"Resp: {r.request.method} {r.status_code} {r.reason} {r.url}")
            if r.raw._connection and r.raw._connection.sock:
                print(f"Conn: {r.raw._connection.sock.getsockname()} -> {r.raw._connection.sock.getpeername()[0]}")
        session.hooks['response'].append(print_request)

    if args.insecure:
        session.verify = False
        requests.packages.urllib3.disable_warnings() # type: ignore
        print("Warning: SSL certificate verification disabled.")
    session_monkey = SessionMonkeyPatch(session=session, delay=args.delay, msg='',
                                        hard_retries=args.hard_retry,
                                        trim_PHP_warnings=args.trim_php_warnings)
    session_monkey.hijack()

    if args.cookies:
        load_cookies(session, args.cookies)
        print("Cookies loaded:", session.cookies.get_dict().keys())

    std_url = standardize_url(url_input)
    # use #force to skip 30X redirection detection
    doku_url = get_doku_url(std_url, session=session) if not std_url.endswith('#force') else std_url[:-len('#force')]

    print('Cookies in use:', session.cookies.get_dict().keys())

    avoidSites(doku_url, session=session)

    if not args.force:
        print("Searching for recent dumps on IA...")
        if any_recent_ia_item_exists(ori_url=doku_url, days=365):
            print("A dump of this wiki was uploaded to IA in the last 365 days. Aborting.")
            sys.exit(88)

    if args.username:
        login_dokuwiki(doku_url=doku_url, session=session,
                       username=args.username, password=args.password)


    base_url = build_base_url(doku_url)
    dump_dir = url2prefix(doku_url) + '-' + \
        time.strftime("%Y%m%d", time.gmtime()) if not args.path else args.path.rstrip('/')
    if args.no_resume:
        if os.path.exists(dump_dir):
            print(
                'Dump directory already exists. (You can use --path to specify a different directory.)')
            return 1

    smkdirs(dump_dir, '/dumpMeta')
    print('Dumping to ', dump_dir,
          '\nBase URL: ', base_url,
          '\nDokuPHP URL: ', doku_url)

    _config = {'url_input': url_input,  # type: str
               'std_url': std_url,  # type: str
               'doku_url': doku_url,  # type: str
               'base_url': base_url,  # type: str
               'dokuWikiDumper_version': get_version(),  # type: str
               }
    update_config(dump_dir=dump_dir, config=_config)
    update_info(dump_dir, doku_url=doku_url, session=session)

    with DumpLock(dump_dir):
        if args.content:
            if os.path.exists(os.path.join(dump_dir, 'content_dumped.mark')):
                print('Content already dumped.')
            else:
                print('\nDumping content...\n')
                dump_content(doku_url=doku_url, dump_dir=dump_dir,
                            session=session, threads=args.threads,
                            ignore_errors=args.ignore_errors,
                            ignore_action_disabled_edit=args.ignore_action_disabled_edit,
                            current_only=args.current_only)
                with open(os.path.join(dump_dir, 'content_dumped.mark'), 'w') as f:
                    f.write('done')
        if args.html:
            if os.path.exists(os.path.join(dump_dir, 'html_dumped.mark')):
                print('HTML already dumped.')
            else:
                print('\nDumping HTML...\n')
                dump_HTML(doku_url=doku_url, dump_dir=dump_dir,
                        session=session, threads=args.threads,
                        ignore_errors=args.ignore_errors, current_only=args.current_only)
                with open(os.path.join(dump_dir, 'html_dumped.mark'), 'w') as f:
                    f.write('done')
        if args.media: # last, so that we can know the dump is complete.
            if os.path.exists(os.path.join(dump_dir, 'media_dumped.mark')):
                print('Media already dumped.')
            else:
                print('\nDumping media...\n')
                dump_media(base_url=base_url, dumpDir=dump_dir,
                        session=session, threads=args.threads,
                        ignore_errors=args.ignore_errors)
                with open(os.path.join(dump_dir, 'media_dumped.mark'), 'w') as f:
                    f.write('done')
        if args.pdf:
            if os.path.exists(os.path.join(dump_dir, 'pdf_dumped.mark')):
                print('PDF already dumped.')
            else:
                print('\nDumping PDF...\n')
                dump_PDF(doku_url=base_url, dump_dir=dump_dir,
                        session=session, threads=args.threads,
                        ignore_errors=args.ignore_errors, current_only=True)
                        # to avoid overload the server, we only dump the current revision of the PDF.
                with open(os.path.join(dump_dir, 'pdf_dumped.mark'), 'w') as f:
                    f.write('done')


    session_monkey.release()
    print('\n\n--Done--')

    if args.upload and args.auto:
        print('Uploading to Internet Archive...')
        from subprocess import call
        time.sleep(5)
        retcode = call([sys.executable, '-m', 'dokuWikiUploader.uploader', dump_dir] + args.uploader_args,
             shell=False, env=os.environ.copy())
        if retcode == 0:
            print('dokuWikiUploader: --upload: Done')
        else:
            print('dokuWikiUploader: --upload: [red] Failed [/red]!!!')
            raise RuntimeError('dokuWikiUploader: --upload: Failed!!!')



import builtins
import os
import re
import sys
import threading
import time
from typing import Any, Optional, List, Union, overload
from urllib.parse import unquote, urlparse, urljoin

from rich import print as rprint
from slugify import slugify
import requests

USE_RICH = True

fileLock = threading.Lock()
printLock = threading.Lock()

@overload
def check_int(s: str)->Optional[str]:...
@overload
def check_int(s: int) -> int: ...
@overload
def check_int(s: float) -> float: ...
@overload
def check_int(s: None) -> None: ...
@overload
def check_int(s: Any) -> Optional[Any]: ...
def check_int(s: Union[int,float,str,None,Any]) -> Union[int,float,str,None]:
    """
    convert intable variable to int if possible

    :return: None if not int-able, else the raw value
    """
    try:
        int(s) # type: ignore
        return s
    except Exception:
        return None

def print_with_lock(*args, **kwargs):
    printLock.acquire()
    if USE_RICH:
        # replace [[ and ]] with "
        try:
            rich_args = args
            rich_args = [re.sub(r'\[\[', '\"', str(arg)) for arg in rich_args]
            rich_args = [re.sub(r'\]\]', '\"', str(arg)) for arg in rich_args]
            rprint(*rich_args, **kwargs)
        except Exception: # fallback to builtins.print
            builtins.print(*args, **kwargs)
    else:
        builtins.print(*args, **kwargs)
    printLock.release()


def _is_disallow_in_robots_txt(robots_txt: str, user_agent: str) -> bool:
    robots_txt = robots_txt.lower()
    user_agent = user_agent.lower()

    lines = robots_txt.splitlines()
    in_relevant_section = False

    for line in lines:
        if line.startswith('user-agent:'):
            current_user_agent = line[len('user-agent:'):].strip()
            in_relevant_section = (current_user_agent == user_agent)
        elif in_relevant_section and line.startswith('disallow:'):
            disallow_path = line[len('disallow:'):].strip()
            if disallow_path == '/':
                return True
    return False

def avoidSites(url: str, session: requests.Session):
    #check robots.txt
    exit_ = False
    try:
        # Don't use the session.get() method here, since we want to avoid the session's retry logic
        r = requests.get(
            urlparse(url).scheme + '://' + urlparse(url).netloc + '/robots.txt',
            cookies=session.cookies, headers=session.headers, verify=session.verify, proxies=session.proxies
        )
        if r.status_code == 200:
            if _is_disallow_in_robots_txt(r.text, 'ia_archiver') or _is_disallow_in_robots_txt(r.text, 'dokuwikidumper'):
                print('This wiki not allow dokuWikiDumper or IA to crawl.')
                exit_ = True
    except Exception as e:
        print('Error: cannot get robots.txt', e)        
    
    if exit_:
        sys.exit(33)

    site = urlparse(url).netloc
    avoidList = ['www.dokuwiki.org']  # TODO: Add more sites
    if site in avoidList:
        if input('\nWarning:\nYou are trying to dump '+site+', which is in the avoid list. \n' +
                 'If you just want to test ' +
                 'if this program can dump dokuwiki successfully, please DO NOT do this, ' +
                 '\nthis will bring a lot of pressure to the server of ' + site +
                 '\n\nContinue anyway? (y/n): ') != 'y':
            sys.exit(1)

        print('You have been warned. :-)')
        time.sleep(3)


def smkdirs(parent: str, *child: str)-> Optional[str]:
    """ safe mkdir, return: True->created, False->existed """
    if parent is None:
        raise ValueError('parent must be specified')
    
    # lstrip / in child
    _child = [c.lstrip('/') for c in child]

    dir = os.path.join(parent, *_child)
    # print(dir)
    fileLock.acquire()
    if not os.path.exists(dir):
        os.makedirs(dir)
        fileLock.release()
        return dir

    fileLock.release()
    return None


def standardize_url(url: str):
    """ 1. unquote url
        2. Add http:// if scheme is missing
        3. Remove port :80 and :443 if http:// and https:// respectively
        4. Convert domain to IDNA
    """
    # TODO: make step 1,2,4 optional and reversible

    url = url.strip()
    url = unquote(url, encoding='utf-8', errors='strict')

    if not url.startswith('http://') and not url.startswith('https://'):
        print('Warning: URL scheme is missing, assuming http://')
        url = 'http://' + url

    Url = urlparse(url)

    assert Url.hostname is not None, 'URL hostname is None'
    idna_hostname = Url.hostname.encode('idna').decode('utf-8')

    if Url.hostname != idna_hostname:
        print('Converting domain to IDNA: ' + Url.hostname + ' -> ' + idna_hostname)
        url = url.replace(Url.hostname, idna_hostname, 1)
    
    if Url.port == 80 and Url.scheme == 'http':
        print('Removing port 80 from URL')
        url = url.replace(':80', '', 1)
    
    if Url.port == 443 and Url.scheme == 'https':
        print('Removing port 443 from URL')
        url = url.replace(':443', '', 1)

    return url


def get_doku_url(url: str, session: requests.Session):
    r = session.get(url)
    parsedUrl = urlparse(r.url)
    dokuUrl = urljoin(
        parsedUrl.scheme + '://' + parsedUrl.netloc + '/',
        parsedUrl.path
    ) # remove query string


    return dokuUrl


def build_base_url(url: str = '') -> str:
    r = urlparse(url)
    path = r.path
    if path and path != '/' and not path.endswith('/'):
        path = path[:path.rfind('/')]
    baseUrl = r.scheme + '://' + r.netloc + path
    if not baseUrl.endswith('/'):
        baseUrl += '/'

    return baseUrl


def url2prefix(url: str, ascii_slugify: bool = True):
    """Convert URL to a valid prefix filename."""

    url = url.strip()
    if '\n' in url or '\r' in url:
        raise ValueError('URL contains newline')
    url = standardize_url(url)

    r = urlparse(url)

    r_path = r.path

    if r.path.endswith('/'):
        # remove last slash
        # "/abc/123/" -> "/abc/123"
        # "/" -> ""
        r_path = r.path[:-1]
    else: # not r.path.endswith('/')
        # truncate to last slash
        # "/abc/123/edf" -> "/abc/123"
        r_path =  r.path[:r.path.rfind('/')]

    # remove "/any.php" suffix
    r_path = re.sub(r"(/[^/]+\.php)", "", r_path)
    # remove tilde
    r_path = r_path.replace('~', '')
    # sulgify
    _r_paths = r_path.split('/')
    if ascii_slugify:
        _r_paths = [slugify(p, separator='_', allow_unicode=False) for p in _r_paths]
    r_path = '_'.join(_r_paths)

    # replace port with underscore
    r_netloc = r.netloc.replace(':', '_')

    # lower case
    prefix = (r_netloc + r_path).lower()
    assert prefix == prefix.strip('_'), 'prefix contains leading or trailing underscore, please report this bug.'

    return prefix

def test_url2prefix():
    # TODO: add test
    pass
    tests_slugify = {
        'https://苟.com:123/~11.22利.国3/家-！!🌚/asd?das#jbcdbd':
        'xn--ui1a.com_123_11_22li_guo_3_jia'
    }
    for url, prefix in tests_slugify.items():
        assert url2prefix(url, ascii_slugify=True) == prefix


def load_titles(titles_file_path) -> Optional[List[str]]:
    """ Load titles from dump directory

    Return:
        `list[str]`: titles
        `None`: titles file does not exist or incomplete
     """
    if os.path.exists(titles_file_path):
        with uopen(titles_file_path, 'r') as f:
            titles = f.read().splitlines()
        if len(titles) and titles[-1] == '--END--':
            print('Loaded %d titles from %s' %
                  (len(titles) - 1, titles_file_path))
            return titles[:-1]  # remove '--END--'

    return None


def uopen(*args, **kwargs):
    """ I dont wanna type `encoding=utf8` anymore.
    Made for Windows compatibility :-( """
    return open(*args, encoding='UTF-8', **kwargs)


WARNINGS_TO_REMOVE = tuple([
    '<br>',
    '<br />',
    '<b>Warning</b>',
    '<b>Error</b>',
    '<b>Notice</b>',
    '<b>Deprecated</b>',
    '<b>Strict Standards</b>',
    '<b>Strict warning</b>',
    '<div class="error">',
    "<div class='error'>",
    '<div class="warning">',
    "<div class='warning'>",
    '<div class="Deprecated">',
    "<div class='Deprecated'>",
    '<error>'
    # add more if needed
])

def trim_PHP_warnings(html_or_text: str, strict: bool = False) -> str:
    """ Trim PHP warnings in HTML or wikitext (Cannot handle single line HTML)
    
    param: `strict`: if `True`, only remove warnings when found `<html` 
    or `<!DOCTYPE html`(return original text if not found).
    """

    doc_type_is_html_TAG = '<!DOCTYPE html' in html_or_text
    html_TAG = '<html' in html_or_text

    if strict and not (doc_type_is_html_TAG or html_TAG):
        return html_or_text

    lines = html_or_text.splitlines(keepends=True) # keepends=True to keep \n

    if len(lines) == 1 and (doc_type_is_html_TAG or html_TAG):
        print('Warning: cannot remove dokuwiki warnings in single line text')
        return html_or_text # original text

    new_text = ''
    in_document = False
    for line in lines:
        if in_document:
            new_text += line
            continue

        if ((doc_type_is_html_TAG and '<!DOCTYPE html' in line) or
            (html_TAG             and '<html'          in line)
        ):
            in_document = True
            new_text = line # first line
            continue
        elif (doc_type_is_html_TAG or html_TAG):
            # remove anything before '<!DOCTYPE html' or '<html'
            print('Removing PHP warning: ' + line.strip())
            continue

        # here: in_document == False, doc_type_is_html_TAG == False, html_TAG == False
        if (line.startswith(WARNINGS_TO_REMOVE)):
            print('(HTML header not found) removing PHP warning (unsafely): ' + line.strip())
            continue
        else:
            in_document = True
            # print('The line is not PHP warning (if it is, please report an issue): ' + line.strip())
            new_text = line # first line
            continue

    def help_wanted(msg: str = ''):
        print_with_lock(msg) # with color
    if html_or_text != new_text:
        help_wanted("[yellow]Notice: trim_PHP_warnings() needs more sample sites to test. "
                "Please feedback if you found some warnings are not removed.[/yellow]")

    return new_text


class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


import threading
import asyncio


delaying_theads = 0
delaying_theads_lock = threading.Lock()

class Delay:
    done: bool = False
    lock: threading.Lock = threading.Lock()
    delay: float

    async def animate(self):
        while True:
            with self.lock:
                if self.done:
                    return

                with delaying_theads_lock:
                    print("\r" + self.ellipses, end="") if delaying_theads == 1 else None
                self.ellipses += "."

            await asyncio.sleep(0.35)


    def __init__(self, msg=None, delay: float=0.0):
        """Add a delay if configured for that"""
        if delay <= 0:
            return
        self.ellipses: str = "."
        self.delay = delay


        with delaying_theads_lock:
            global delaying_theads
            delaying_theads += 1
            
        if msg:
            self.ellipses = ("Delay %.1fs: %s " % (delay, msg)) + self.ellipses
        else:
            self.ellipses = ("Delay %.1fs " % (delay)) + self.ellipses

        asyncio.run(self.async_tasks())

    async def async_tasks(self):
        task1 = asyncio.create_task(self.animate())
        task2 = asyncio.create_task(self.task_done())
        await task2
        task1.cancel()

    async def task_done(self):
        await asyncio.sleep(self.delay)

        with self.lock and delaying_theads_lock:
            self.done = True
            global delaying_theads
            delaying_theads -= 1
            if delaying_theads == 0: # If this is the last thread, clear the line
                print("\r" + " " * len(self.ellipses) + "\r", end="")



import json
import os
import re
import base64
from typing import Optional, Union
from urllib.parse import urljoin

from bs4 import BeautifulSoup
from bs4.element import Tag
import requests

from dokuWikiDumper.utils.util import print_with_lock as print
from dokuWikiDumper.utils.util import uopen
from dokuWikiDumper.utils.config import runtime_config


INFO_FILEPATH = 'dumpMeta/info.json'
HOMEPAGE_FILEPATH = 'dumpMeta/index.html'
CHECKPAGE_FILEPATH = 'dumpMeta/check.html'
ICON_FILEPATH = 'dumpMeta/favicon.ico'

# info.json keys
INFO_WIKI_NAME = "wiki_name"
INFO_RAW_TITLE = "raw_title"
INFO_DOKU_URL = "doku_url"
INFO_LANG = "language"
INFO_ICON_URL = "icon_url"

def get_info(dumpDir: str) -> dict:
    if os.path.exists(os.path.join(dumpDir, INFO_FILEPATH)):
        with uopen(os.path.join(dumpDir, INFO_FILEPATH), 'r') as f:
            _info = json.load(f)
            return _info

    return {}


def update_info_json(dumpDir: str, info: dict):
    '''Only updates given keys in info.'''
    _info = get_info(dumpDir)
    info = {**_info, **info}

    with uopen(os.path.join(dumpDir, INFO_FILEPATH), 'w') as f:
        json.dump(info, f, indent=4, ensure_ascii=False)


def get_html_lang(html: Union[bytes, str]) -> Optional[str]:
    '''Returns the language of the html document.'''

    soup = BeautifulSoup(html, runtime_config.html_parser)
    # <html lang="en" dir="ltr" class="no-js">
    html_tag = soup.html
    
    lang = None
    if html_tag:
        lang = html_tag.get('lang')
    
    if isinstance(lang, list):
        lang = lang[0]

    return lang


def get_wiki_name(html: Union[bytes, str]):
    '''Returns the name of the wiki.

    Tuple: (wiki_name: Optional[str], raw_title: Optional[str])'''

    soup = BeautifulSoup(html, runtime_config.html_parser)
    try:
        raw_title = soup.title.text
    except:
        print(    '[red]Error: Could not find HTML title[/red]', end='')
        if isinstance(html, str) and 'Warning' in html:
            print(' [red]and the HTML contains PHP warning.[/red]')
            print('       Re-run with --trim-php-warnings or change the --parser may help.')
        else:
            print('.')
        raise
    wiki_name = re.search(r'\[(.+)\]', raw_title)  # 'start [wikiname]'.
    if wiki_name is not None:
        wiki_name = wiki_name.group(1)
    else:
        print('Warning: Could not find wiki name in HTML title.')
    
    if not isinstance(wiki_name, (str, type(None))):
        wiki_name = None

    if isinstance(wiki_name, str):
        wiki_name = wiki_name.replace('\n', ' ').replace('\r', '').strip()

    return wiki_name, raw_title


def get_raw_icon_href(html: Union[bytes ,str]):
    '''Returns the icon url.'''

    soup = BeautifulSoup(html, runtime_config.html_parser)
    icon_tag = soup.find('link', rel='shortcut icon')
    icon_url = None
    if isinstance(icon_tag, Tag):
        icon_url = icon_tag.get('href')
        if isinstance(icon_url, list):
            icon_url = icon_url[0]
    else:
        print('Warning: Could not find icon in HTML.')

    return icon_url


def save_icon(dumpDir: str, url_or_dataUrl: str, session: requests.Session):
    if url_or_dataUrl is None:
        return False
    if url_or_dataUrl.startswith('data:image'):
        assert 'base64,' in url_or_dataUrl, 'Icon is not base64 encoded.'
        print('Saving icon from data url. (base64 encoded)')
        data_base64 = url_or_dataUrl.split('base64,')[1]
        data = base64.urlsafe_b64decode(data_base64)
        with open(os.path.join(dumpDir, ICON_FILEPATH), 'wb') as f:
            f.write(data)
        return True

    r = session.get(url_or_dataUrl)
    if 'image' not in r.headers.get('content-type', ''):
        print('Warning: Icon is not an image.')
        return False
    with open(os.path.join(dumpDir, ICON_FILEPATH), 'wb') as f:
        f.write(r.content)
        return True



def update_info(dumpDir: str, doku_url: str, session: requests.Session):
    '''Saves the info of the wiki.'''
    homepage_html = session.get(doku_url).text
    with uopen(os.path.join(dumpDir, HOMEPAGE_FILEPATH), 'w') as f:
        f.write(homepage_html)
        print('Saved homepage to', HOMEPAGE_FILEPATH)

    checkpage_html = session.get(doku_url, params={'do': 'check'}).content
    with open(os.path.join(dumpDir, CHECKPAGE_FILEPATH), 'wb') as f:
        f.write(checkpage_html)
        print('Saved checkpage to', CHECKPAGE_FILEPATH)

    wiki_name, raw_title = get_wiki_name(homepage_html)
    lang = get_html_lang(homepage_html)
    icon_href = get_raw_icon_href(homepage_html)
    icon_url = None
    if icon_href is not None and icon_href.startswith('data:image'):
        # base64 encoded data url
        icon_url = icon_href
    elif icon_href is not None:
        # normal url
        icon_url = urljoin(doku_url, icon_href)

    save_icon(dumpDir=dumpDir, url_or_dataUrl=icon_url, session=session)

    info = {
        INFO_WIKI_NAME: wiki_name,
        INFO_RAW_TITLE: raw_title,
        INFO_DOKU_URL: doku_url,
        INFO_LANG: lang,
        INFO_ICON_URL: icon_url,
    }
    print('Info:', info)
    update_info_json(dumpDir, info)
